(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: og_config.ml,v 1.6 2014/01/31 09:51:52 fclement Exp $ *)

let config_error message =
  failwith (Printf.sprintf "%s: %s" Sys.executable_name message)
;;

let get_keep_comments, set_keep_comments =
  let keep_comments = ref false in
  (fun () -> !keep_comments),
  (fun b -> keep_comments := b)
;;

let get_keep_type_definitions, set_keep_type_definitions =
  let keep_types = ref false in
  (fun () -> !keep_types),
  (fun b -> keep_types := b)
;;

let get_ignore_explicit_alias, set_ignore_explicit_alias =
  let ignore_explicit_alias = ref true in
  (fun () -> !ignore_explicit_alias),
  (fun b -> ignore_explicit_alias := b)
;;

let set_input_file_name, get_input_file_name, is_set_input_file_name,
    set_output_file_name, get_output_file_name, is_set_output_file_name =

  let src = ref None in
  let tgt = ref None in

  let set_src src_name =
    let tgt_name =
      match !tgt with
      | None -> "-"
      | Some tgt_name -> tgt_name in
    if Og_file.safe_source_file_processing src_name tgt_name
    then src := Some src_name 
    else
      config_error (Printf.sprintf
        "target file %s would overwrite source file %s"
        tgt_name src_name) in

  let set_tgt tgt_name =
    let src_name =
      match !src with
      | None -> "-"
      | Some src_name -> src_name in
    if Og_file.safe_source_file_processing src_name tgt_name
    then tgt := Some tgt_name 
    else
      config_error (Printf.sprintf
        "target file %s would overwrite source file %s"
        tgt_name src_name) in

  (fun s ->
   match !src with
   | None -> set_src s
   | Some fname ->
     config_error (Printf.sprintf
       "can not set input file name to %s, it is already set to %s"
       s fname)
  ),
  (fun () ->
   match !src with
   | None ->
     config_error "input file is not set"
   | Some fname -> fname
  ),
  (fun () ->
   match !src with
   | None -> false
   | Some _ -> true
  ),

  (fun s ->
   match !tgt with
   | None -> set_tgt s
   | Some fname ->
     config_error (Printf.sprintf
       "can not set output file name to %s, it is already set to %s"
        s fname)
  ),
  (fun () ->
   match !tgt with
   | None ->
     config_error "output file is not set"
   | Some fname -> fname
  ),
  (fun () ->
   match !tgt with
   | None -> false
   | Some _ -> true
  )
;;

let set_anonymous_file_name, get_anonymous_file_name =

  let anonymous_fname = ref None in

  (fun suffix s ->
   match !anonymous_fname with
   | None ->
     anonymous_fname := Some s;
     if is_set_input_file_name ()
     then
       if is_set_output_file_name ()
       then
         config_error (Printf.sprintf
           "can not set anonymous file to %s, \
            input and output files are already set"
           s)
       else
         set_output_file_name
           (Og_file.add_suffix_extension suffix ".ml" s)
     else
     begin
       set_input_file_name s;
       if not (is_set_output_file_name ()) then
         set_output_file_name
           (Og_file.add_suffix_extension suffix ".ml" s)
     end
   | Some fname ->
     config_error (Printf.sprintf
       "can not set anonymous file argument to %s, it is already set to %s"
        s fname)
  ),
  (fun () -> !anonymous_fname)
;;

let get_input_output_file_names () =
  (get_input_file_name ()), (get_output_file_name ())
;;

let get_input_module_name () =
  let module_file_name = Filename.basename (get_input_file_name ()) in
  String.capitalize (Og_file.chop_extension module_file_name)
;;

let print_full_version () =
  prerr_endline Og_version.full;
  exit 0
;;

let print_short_version () =
  prerr_endline Og_version.short;
  exit 0
;;
