(***********************************************************************)
(*                                                                     *)
(*                              OCamlGen                               *)
(*                                                                     *)
(*                       OCaml module Generator                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 1997-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: walker.ml,v 1.17 2014/01/31 09:51:52 fclement Exp $ *)

open Walker_ast;;

(*
   The functor that generates functions over a type:
   walk the type, calling functions from module S to write patterns of
   walking functions,
   calling functions from module U to write expressions of walking
   functions.
*)

module Make
         (S0 : Walk_spec.Make)
         (U0 : Action_spec.Make)
         (Io : Walk_spec.Out) =
struct

  module S = S0 (Io);;
  module U = U0 (Io);;

  let rec pattern = function
    | Pvariable (s, _ty) ->
      S.variable s
    | Ptuple (pa, _ty) ->
      S.begin_tuple ();
      pattern pa.(0);
      for i = 1 to Array.length pa - 1 do
        S.separator_tuple ();
        pattern pa.(i)
      done;
      S.end_tuple ()
  ;;

  let buffer_argument = function
    | None -> ()
    | Some pa ->
      S.separator_argument ();
      pattern pa
  ;;

  let function_name_name = function
      (* Argument of a functional walk *)
    | Fname_arg s ->
      Og_debug.message (Printf.sprintf "*********Fname_arg %s **********" s);
      S.variable s
      (* Named walk *)
    | Fname_local s ->
      Og_debug.message (Printf.sprintf "*********Fname_local %s **********" s);
      S.variable s
      (* External walk *)
    | Fname_external s ->
      Og_debug.message (Printf.sprintf "*********Fname_external %s **********" s);
      U.external_fun s
  ;;

  let rec function_expr = function
    | Fconstant f_name ->
      Og_debug.message (Printf.sprintf "*********Fconstant **********");
      function_name_name f_name
    | Fapply (f_name, function_expr_list) ->
      Og_debug.message (Printf.sprintf "*********Fapply **********");
      S.begin_application ();
        function_name_name f_name;
        List.iter
          (fun f ->
            S.separator_argument ();
            function_expr f)
          function_expr_list;
      S.end_application ()
    | Fabs (patl, e) ->
      Og_debug.message (Printf.sprintf "*********Fabs **********");
      S.begin_function ();
        S.begin_clause_pattern ();
          Og_misc.iter_sep
            (fun _ -> S.separator_argument ())
            pattern
            patl;
        S.end_clause_pattern ();
        S.begin_clause_expr ();
          expr e;
        S.end_clause_expr ();
      S.end_function ()
    | Fparen f ->
      S.begin_paren ();
        function_expr f;
      S.end_paren ()

  and expr = function
(*    | Econstant f -> function_expr f *)
      (* Walk application *)
    | Eapply (f, arg_list) ->
      let argument = function
        | Avar pat -> pattern pat
        | Ainj (inj_name, pat) ->
          S.begin_paren ();
            function_name_name inj_name;
            S.separator_argument ();
            pattern pat;
          S.end_paren () in

      U.begin_application ();
        function_expr f;
        List.iter
          (fun arg ->
           U.separator_argument ();
           U.begin_argument ();
             argument arg;
           U.end_argument ())
          arg_list;
      U.end_application ()
      (* Tuples printing *)
    | Etuple ea ->
      U.begin_tuple ();
        let walk_tuple_elem e =
          U.begin_tuple_elem ();
            expr e;
          U.end_tuple_elem () in
        walk_tuple_elem ea.(0);
        for i = 1 to Array.length ea - 1 do
          U.separator_tuple ();
          walk_tuple_elem ea.(i)
        done;
      U.end_tuple ()
  ;;

  let sum_field _vis = function
    | Sum_tag sum_label ->
      S.begin_clause_pattern ();
        S.sum_tag sum_label;
      S.end_clause_pattern ();
      S.begin_clause_expr ();
        U.sum_tag sum_label;
      S.end_clause_expr ();
    | Sum_inj (sum_label, Clause (pat, e)) ->
      S.begin_clause_pattern ();
        S.begin_sum_inj sum_label;
          pattern pat;
        S.end_sum_inj sum_label;
      S.end_clause_pattern ();
      S.begin_clause_expr ();
        U.begin_sum_inj sum_label;
          expr e;
        U.end_sum_inj sum_label;
      S.end_clause_expr ();
  ;;

  let sum_clause vis sum_fields =
    List.iter
      (fun field ->
       S.begin_sum_field ();
         sum_field vis field;
       S.end_sum_field ())
    sum_fields
  ;;

  let alias_clause _vis (Clause (pat, e)) =
    S.begin_aliased_clause ();
    S.begin_clause_pattern ();
      pattern pat;
    S.end_clause_pattern ();
    S.begin_clause_expr ();
      expr e;
    S.end_clause_expr ();
    S.end_aliased_clause ();
  ;;

  let record_field_pat _vis = function
    | Record_field (record_label, Clause (pat, _e)) ->
      S.begin_record_field record_label;
        pattern pat;
      S.end_record_field ()
  ;;

  let record_field_expr _vis = function
    | Record_field (record_label, Clause (_pat, e)) ->
      U.begin_record_field record_label;
        expr e;
      U.end_record_field record_label
  ;;

  let record_clause vis (record_fields as _clause) =
    S.begin_record_pattern ();
    List.iter
      (function rf ->
       record_field_pat vis rf;
       S.separator_record_field ())
      record_fields;
    S.end_record_pattern ();
    S.begin_clause_expr ();
      U.begin_record_expr ();
        List.iter (record_field_expr vis) record_fields;
      U.end_record_expr ();
    S.end_clause_expr ();
  ;;

  let rec kind = function
    | Record (b, vis, clause) ->
      buffer_argument b;
      S.begin_record_clause ();
      record_clause vis clause;
      S.end_record_clause ()
    | Sum (b, vis, clause) ->
      buffer_argument b;
      S.begin_sum_clause ();
      sum_clause vis clause;
      S.end_sum_clause ()
    | Alias (b, vis, (cls, None)) ->
      buffer_argument b;
      S.begin_alias_clause ();
      alias_clause vis cls;
      S.end_alias_clause ()
    | Alias (b, vis, (cls, Some ty_kind)) ->
      if Og_config.get_ignore_explicit_alias ()
      then kind (Alias (b, vis, (cls, None)))
      else kind ty_kind
    | External (_, _) -> assert false
  ;;

  let walk_def_elem ((ty_name, walker_name), ty_arguments, ty_kind) =
    S.begin_def_elem ((ty_name, walker_name), ty_arguments);
      kind ty_kind;
    S.end_def_elem ()
  ;;

  let walk_dir = function
    | Dir_open m -> S.directive (Printf.sprintf "open %s" m)
    | Dir_include m -> S.comment (Printf.sprintf "(* include %s *)" m)
  ;;

  let phrase = function
    | Type_definition (rec_flag, ty_def_list) ->
      begin match ty_def_list with
      | [] -> ()
      | ty_def :: ty_def_list ->
        S.begin_def rec_flag;
          walk_def_elem ty_def;
          List.iter
            (function d -> S.separator_def_elem (); walk_def_elem d)
            ty_def_list;
        S.end_def ()
      end
    | Comment s -> S.comment s
    | Directive d -> walk_dir d
  ;;

  let file phs =
    S.begin_file ();
      List.iter phrase phs;
    S.end_file ();
  ;;

end
;;
